@@ -2,6 +2,8 @@
 
 #include <math.h>
 #include <string.h>
+#include <stdatomic.h>
+
 #include <girara/datastructures.h>
 #include <girara/utils.h>
 
@@ -9,30 +11,19 @@
 #include "adjustment.h"
 #include "zathura.h"
 #include "document.h"
+#include "document-widget.h"
 #include "page.h"
 #include "page-widget.h"
 #include "utils.h"
 
 /* private data for ZathuraRenderer */
 typedef struct private_s {
-  GThreadPool* pool;            /**< Pool of threads */
-  GMutex mutex;                 /**< Render lock */
-  volatile bool about_to_close; /**< Render thread is to be freed */
+  GThreadPool* pool;       /**< Pool of threads */
+  girara_list_t* requests; /**< Render requests */
+  GMutex mutex;            /**< Render lock */
 
   /**
-   * recolor information
-   */
-  struct {
-    bool enabled;
-    bool hue;
-    bool reverse_video;
-
-    GdkRGBA light;
-    GdkRGBA dark;
-  } recolor;
-
-  /*
-   * page cache
+   * Page cache
    */
   struct {
     int* cache;
@@ -40,8 +31,19 @@ typedef struct private_s {
     size_t num_cached_pages;
   } page_cache;
 
-  /* render requests */
-  girara_list_t* requests;
+  /**
+   * Recolor information
+   */
+  struct {
+    GdkRGBA light;
+    GdkRGBA dark;
+    bool enabled;
+    bool hue;
+    bool reverse_video;
+    bool adjust_lightness;
+  } recolor;
+
+  atomic_bool about_to_close; /**< Render thread is to be freed */
 } ZathuraRendererPrivate;
 
 /* private data for ZathuraRenderRequest */
@@ -74,7 +76,7 @@ static bool page_cache_is_full(ZathuraRenderer* renderer, bool* result);
 /* job descritption for render thread */
 typedef struct render_job_s {
   ZathuraRenderRequest* request;
-  volatile bool aborted;
+  atomic_bool aborted;
 } render_job_t;
 
 /* init, new and free for ZathuraRenderer */
@@ -93,9 +95,10 @@ static void zathura_renderer_init(ZathuraRenderer* renderer) {
   g_mutex_init(&priv->mutex);
 
   /* recolor */
-  priv->recolor.enabled       = false;
-  priv->recolor.hue           = true;
-  priv->recolor.reverse_video = false;
+  priv->recolor.enabled          = false;
+  priv->recolor.hue              = true;
+  priv->recolor.reverse_video    = false;
+  priv->recolor.adjust_lightness = false;
 
   /* page cache */
   priv->page_cache.size             = 0;
@@ -291,6 +294,21 @@ void zathura_renderer_enable_recolor_reverse_video(ZathuraRenderer* renderer, bo
   ZathuraRendererPrivate* priv = zathura_renderer_get_instance_private(renderer);
   priv->recolor.reverse_video  = enable;
 }
+
+bool zathura_renderer_recolor_adjust_lightness_enabled(ZathuraRenderer* renderer) {
+  g_return_val_if_fail(ZATHURA_IS_RENDERER(renderer), false);
+
+  ZathuraRendererPrivate* priv = zathura_renderer_get_instance_private(renderer);
+  return priv->recolor.adjust_lightness;
+}
+
+void zathura_renderer_enable_recolor_adjust_lightness(ZathuraRenderer* renderer, bool enable) {
+  g_return_if_fail(ZATHURA_IS_RENDERER(renderer));
+
+  ZathuraRendererPrivate* priv   = zathura_renderer_get_instance_private(renderer);
+  priv->recolor.adjust_lightness = enable;
+}
+
 void zathura_renderer_set_recolor_colors(ZathuraRenderer* renderer, const GdkRGBA* light, const GdkRGBA* dark) {
   g_return_if_fail(ZATHURA_IS_RENDERER(renderer));
 
@@ -325,10 +343,10 @@ void zathura_renderer_get_recolor_colors(ZathuraRenderer* renderer, GdkRGBA* lig
 
   ZathuraRendererPrivate* priv = zathura_renderer_get_instance_private(renderer);
   if (light != NULL) {
-    memcpy(light, &priv->recolor.light, sizeof(GdkRGBA));
+    *light = priv->recolor.light;
   }
   if (dark != NULL) {
-    memcpy(dark, &priv->recolor.dark, sizeof(GdkRGBA));
+    *dark = priv->recolor.dark;
   }
 }
 
@@ -452,6 +470,17 @@ static gboolean emit_completed_signal(void* data) {
   return FALSE;
 }
 
+static bool pixel_inside_rectangles(girara_list_t* rectangles, unsigned int x, unsigned int y) {
+  for (size_t idx = 0; idx != girara_list_size(rectangles); ++idx) {
+    zathura_rectangle_t* rect_it = girara_list_nth(rectangles, idx);
+    if (rect_it->x1 <= x && rect_it->x2 >= x && rect_it->y1 <= y && rect_it->y2 >= y) {
+      return true;
+    }
+  }
+
+  return false;
+}
+
 /* Returns the maximum possible saturation for given h and l.
    Assumes that l is in the interval l1, l2 and corrects the value to
    force u=0 on l1 and l2 */
@@ -480,30 +509,11 @@ static double colorumax(const double h[3], double l, double l1, double l2) {
   return fmin(u, v);
 }
 
-static void recolor(ZathuraRendererPrivate* priv, zathura_page_t* page, unsigned int page_width,
-                    unsigned int page_height, cairo_surface_t* surface) {
-  /* uses a representation of a rgb color as follows:
-     - a lightness scalar (between 0,1), which is a weighted average of r, g, b,
-     - a hue vector, which indicates a radian direction from the grey axis,
-       inside the equal lightness plane.
-     - a saturation scalar between 0,1. It is 0 when grey, 1 when the color is
-       in the boundary of the rgb cube.
-  */
-
-  /* TODO: split handling of image handling off
-   * Ideally we would create a mask surface for the location of the images and
-   * we would blit the recolored and unmodified surfaces together to get the
-   * same effect.
-   */
-
-  cairo_surface_flush(surface);
-
-  const int rowstride  = cairo_image_surface_get_stride(surface);
-  unsigned char* image = cairo_image_surface_get_data(surface);
-
-  /* RGB weights for computing lightness. Must sum to one */
-  static const double weights[] = {0.30, 0.59, 0.11};
+/* RGB weights for computing lightness. Must sum to one */
+static const double weights[] = {0.30, 0.59, 0.11};
 
+static void recolor_slow(ZathuraRendererPrivate* priv, unsigned int page_width, unsigned int page_height,
+                         cairo_surface_t* surface, girara_list_t* rectangles, bool found_images) {
   const GdkRGBA rgb1 = priv->recolor.dark;
   const GdkRGBA rgb2 = priv->recolor.light;
 
@@ -526,40 +536,10 @@ static void recolor(ZathuraRendererPrivate* priv, zathura_page_t* page, unsigned
       rgb2.blue * rgb2.alpha - l2,
   };
 
-  /* Decide if we can use the older, faster formulas */
-  const bool fast_formula =
-      (!priv->recolor.hue || (fabs(rgb1.red - rgb1.blue) < DBL_EPSILON && fabs(rgb1.red - rgb1.green) < DBL_EPSILON &&
-                              fabs(rgb2.red - rgb2.blue) < DBL_EPSILON && fabs(rgb2.red - rgb2.green) < DBL_EPSILON)) &&
-      (rgb1.alpha >= 1. - DBL_EPSILON && rgb2.alpha >= 1. - DBL_EPSILON);
+  bool adjust_lightness = priv->recolor.adjust_lightness;
 
-  girara_list_t* images     = NULL;
-  girara_list_t* rectangles = NULL;
-  bool found_images         = false;
-
-  /* If in reverse video mode retrieve images */
-  if (priv->recolor.reverse_video == true) {
-    images       = zathura_page_images_get(page, NULL);
-    found_images = (images != NULL);
-
-    rectangles = girara_list_new();
-    if (rectangles == NULL) {
-      found_images = false;
-      girara_warning("Failed to retrieve images.");
-    }
-
-    if (found_images == true) {
-      /* Get images bounding boxes */
-      for (size_t idx = 0; idx != girara_list_size(images); ++idx) {
-        zathura_image_t* image_it = girara_list_nth(images, idx);
-        zathura_rectangle_t* rect = g_try_malloc(sizeof(zathura_rectangle_t));
-        if (rect == NULL) {
-          break;
-        }
-        *rect = recalc_rectangle(page, image_it->position);
-        girara_list_append(rectangles, rect);
-      }
-    }
-  }
+  const int rowstride  = cairo_image_surface_get_stride(surface);
+  unsigned char* image = cairo_image_surface_get_data(surface);
 
   for (unsigned int y = 0; y < page_height; y++) {
     unsigned char* data = image + y * rowstride;
@@ -567,14 +547,7 @@ static void recolor(ZathuraRendererPrivate* priv, zathura_page_t* page, unsigned
     for (unsigned int x = 0; x < page_width; x++, data += 4) {
       /* Check if the pixel belongs to an image when in reverse video mode*/
       if (priv->recolor.reverse_video == true && found_images == true) {
-        bool inside_image = false;
-        for (size_t idx = 0; idx != girara_list_size(rectangles); ++idx) {
-          zathura_rectangle_t* rect_it = girara_list_nth(rectangles, idx);
-          if (rect_it->x1 <= x && rect_it->x2 >= x && rect_it->y1 <= y && rect_it->y2 >= y) {
-            inside_image = true;
-            break;
-          }
-        }
+        const bool inside_image = pixel_inside_rectangles(rectangles, x, y);
         /* If it's inside and image don't recolor */
         if (inside_image == true) {
           /* It is not guaranteed that the pixel is already opaque. */
@@ -599,52 +572,192 @@ static void recolor(ZathuraRendererPrivate* priv, zathura_page_t* page, unsigned
         const double u = colorumax(h, l, 0, 1);
         const double s = fabs(u) > DBL_EPSILON ? 1.0 / u : 0.0;
 
+        /* adjust according to quartic curve, then average with original weighed
+         * by half saturation. */
+        if (adjust_lightness) {
+          /* l = l * s/2 + l^4 * (1 - s/2) */
+          double adj = l * l * l * l;
+          l          = (l - adj) * (s * 0.5) + adj;
+        }
+
         /* Interpolates lightness between light and dark colors. white goes to
          * light, and black goes to dark. */
         l = l * (l2 - l1) + l1;
 
         const double su = s * colorumax(h, l, l1, l2);
 
-        if (fast_formula) {
-          data[3] = 255;
-          data[2] = (unsigned char)round(255. * (l + su * h[0]));
-          data[1] = (unsigned char)round(255. * (l + su * h[1]));
-          data[0] = (unsigned char)round(255. * (l + su * h[2]));
-        } else {
-          /* Mix lightcolor, darkcolor and the original color, according to the
-           * minimal and maximal channel of the original color */
-          const double tr1 = (1. - fmax(fmax(rgb[0], rgb[1]), rgb[2]));
-          const double tr2 = fmin(fmin(rgb[0], rgb[1]), rgb[2]);
-          data[3]          = (unsigned char)round(255. * (1. - tr1 * negalph1 - tr2 * negalph2));
-          data[2]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[0] + tr2 * h2[0] + (l + su * h[0]))));
-          data[1]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[1] + tr2 * h2[1] + (l + su * h[1]))));
-          data[0]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[2] + tr2 * h2[2] + (l + su * h[2]))));
-        }
+        /* Mix lightcolor, darkcolor and the original color, according to the
+         * minimal and maximal channel of the original color */
+        const double tr1 = (1. - fmax(fmax(rgb[0], rgb[1]), rgb[2]));
+        const double tr2 = fmin(fmin(rgb[0], rgb[1]), rgb[2]);
+        data[3]          = (unsigned char)round(255. * (1. - tr1 * negalph1 - tr2 * negalph2));
+        data[2]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[0] + tr2 * h2[0] + (l + su * h[0]))));
+        data[1]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[1] + tr2 * h2[1] + (l + su * h[1]))));
+        data[0]          = (unsigned char)round(255. * fmin(1, fmax(0, tr1 * h1[2] + tr2 * h2[2] + (l + su * h[2]))));
       } else {
+        if (adjust_lightness) {
+          l = l * l;
+        }
+
         /* linear interpolation between dark and light with color ligtness as
          * a parameter */
-        if (fast_formula) {
+        const double f1 = 1. - (1. - fmax(fmax(rgb[0], rgb[1]), rgb[2])) * negalph1;
+        const double f2 = fmin(fmin(rgb[0], rgb[1]), rgb[2]) * negalph2;
+        data[3]         = (unsigned char)round(255. * (f1 - f2));
+        data[2]         = (unsigned char)round(255. * (l * rgb_diff[0] - f2 * rgb2.red + f1 * rgb1.red));
+        data[1]         = (unsigned char)round(255. * (l * rgb_diff[1] - f2 * rgb2.green + f1 * rgb1.green));
+        data[0]         = (unsigned char)round(255. * (l * rgb_diff[2] - f2 * rgb2.blue + f1 * rgb1.blue));
+      }
+    }
+  }
+}
+
+static void recolor_fast(ZathuraRendererPrivate* priv, unsigned int page_width, unsigned int page_height,
+                         cairo_surface_t* surface, girara_list_t* rectangles, bool found_images) {
+  const GdkRGBA rgb1 = priv->recolor.dark;
+  const GdkRGBA rgb2 = priv->recolor.light;
+
+  const double l1 = weights[0] * rgb1.red + weights[1] * rgb1.green + weights[2] * rgb1.blue;
+  const double l2 = weights[0] * rgb2.red + weights[1] * rgb2.green + weights[2] * rgb2.blue;
+
+  const double rgb_diff[] = {rgb2.red - rgb1.red, rgb2.green - rgb1.green, rgb2.blue - rgb1.blue};
+
+  bool adjust_lightness = priv->recolor.adjust_lightness;
+
+  const int rowstride  = cairo_image_surface_get_stride(surface);
+  unsigned char* image = cairo_image_surface_get_data(surface);
+
+  for (unsigned int y = 0; y < page_height; y++) {
+    unsigned char* data = image + y * rowstride;
+
+    for (unsigned int x = 0; x < page_width; x++, data += 4) {
+      /* Check if the pixel belongs to an image when in reverse video mode*/
+      if (priv->recolor.reverse_video == true && found_images == true) {
+        const bool inside_image = pixel_inside_rectangles(rectangles, x, y);
+        /* If it's inside and image don't recolor */
+        if (inside_image == true) {
+          /* It is not guaranteed that the pixel is already opaque. */
           data[3] = 255;
-          data[2] = (unsigned char)round(255. * (l * rgb_diff[0] + rgb1.red));
-          data[1] = (unsigned char)round(255. * (l * rgb_diff[1] + rgb1.green));
-          data[0] = (unsigned char)round(255. * (l * rgb_diff[2] + rgb1.blue));
-        } else {
-          const double f1 = 1. - (1. - fmax(fmax(rgb[0], rgb[1]), rgb[2])) * negalph1;
-          const double f2 = fmin(fmin(rgb[0], rgb[1]), rgb[2]) * negalph2;
-          data[3]         = (unsigned char)round(255. * (f1 - f2));
-          data[2]         = (unsigned char)round(255. * (l * rgb_diff[0] - f2 * rgb2.red + f1 * rgb1.red));
-          data[1]         = (unsigned char)round(255. * (l * rgb_diff[1] - f2 * rgb2.green + f1 * rgb1.green));
-          data[0]         = (unsigned char)round(255. * (l * rgb_diff[2] - f2 * rgb2.blue + f1 * rgb1.blue));
+          continue;
         }
       }
+
+      /* Careful. data color components blue, green, red. */
+      const double rgb[3] = {data[2] / 255., data[1] / 255., data[0] / 255.};
+
+      /* compute h, s, l data   */
+      double l = weights[0] * rgb[0] + weights[1] * rgb[1] + weights[2] * rgb[2];
+
+      if (priv->recolor.hue == true) {
+        /* adjusting lightness keeping hue of current color. white and black
+         * go to grays of same ligtness as light and dark colors. */
+        const double h[3] = {rgb[0] - l, rgb[1] - l, rgb[2] - l};
+
+        /* u is the maximum possible saturation for given h and l. s is a
+         * rescaled saturation between 0 and 1 */
+        const double u = colorumax(h, l, 0, 1);
+        const double s = fabs(u) > DBL_EPSILON ? 1.0 / u : 0.0;
+
+        /* adjust according to quartic curve, then average with original weighed
+         * by half saturation. */
+        if (adjust_lightness) {
+          /* l = l * s/2 + l^4 * (1 - s/2) */
+          double adj = l * l * l * l;
+          l          = (l - adj) * (s * 0.5) + adj;
+        }
+
+        /* Interpolates lightness between light and dark colors. white goes to
+         * light, and black goes to dark. */
+        l = l * (l2 - l1) + l1;
+
+        const double su = s * colorumax(h, l, l1, l2);
+
+        /* Mix lightcolor, darkcolor and the original color, according to the
+         * minimal and maximal channel of the original color */
+        data[3] = 255;
+        data[2] = (unsigned char)round(255. * (l + su * h[0]));
+        data[1] = (unsigned char)round(255. * (l + su * h[1]));
+        data[0] = (unsigned char)round(255. * (l + su * h[2]));
+      } else {
+        if (adjust_lightness) {
+          l = l * l;
+        }
+
+        /* linear interpolation between dark and light with color ligtness as
+         * a parameter */
+        data[3] = 255;
+        data[2] = (unsigned char)round(255. * (l * rgb_diff[0] + rgb1.red));
+        data[1] = (unsigned char)round(255. * (l * rgb_diff[1] + rgb1.green));
+        data[0] = (unsigned char)round(255. * (l * rgb_diff[2] + rgb1.blue));
+      }
+    }
+  }
+}
+
+static void recolor(ZathuraRendererPrivate* priv, zathura_page_t* page, unsigned int page_width,
+                    unsigned int page_height, cairo_surface_t* surface, zathura_device_factors_t device_factors) {
+  /* uses a representation of a rgb color as follows:
+     - a lightness scalar (between 0,1), which is a weighted average of r, g, b,
+     - a hue vector, which indicates a radian direction from the grey axis,
+       inside the equal lightness plane.
+     - a saturation scalar between 0,1. It is 0 when grey, 1 when the color is
+       in the boundary of the rgb cube.
+  */
+
+  /* TODO: split handling of image handling off
+   * Ideally we would create a mask surface for the location of the images and
+   * we would blit the recolored and unmodified surfaces together to get the
+   * same effect.
+   */
+
+  cairo_surface_flush(surface);
+
+  const GdkRGBA rgb1 = priv->recolor.dark;
+  const GdkRGBA rgb2 = priv->recolor.light;
+
+  /* Decide if we can use the older, faster formulas */
+  const bool fast_formula =
+      (!priv->recolor.hue || (fabs(rgb1.red - rgb1.blue) < DBL_EPSILON && fabs(rgb1.red - rgb1.green) < DBL_EPSILON &&
+                              fabs(rgb2.red - rgb2.blue) < DBL_EPSILON && fabs(rgb2.red - rgb2.green) < DBL_EPSILON)) &&
+      (rgb1.alpha >= 1. - DBL_EPSILON && rgb2.alpha >= 1. - DBL_EPSILON);
+
+  g_autoptr(girara_list_t) rectangles = NULL;
+  bool found_images                   = false;
+
+  /* If in reverse video mode retrieve images */
+  if (priv->recolor.reverse_video == true) {
+    g_autoptr(girara_list_t) images = zathura_page_images_get(page, NULL);
+    found_images                    = (images != NULL);
+
+    rectangles = girara_list_new();
+    if (rectangles == NULL) {
+      found_images = false;
+      girara_warning("Failed to retrieve images.");
+    }
+
+    if (found_images == true) {
+      /* Get images bounding boxes */
+      for (size_t idx = 0; idx != girara_list_size(images); ++idx) {
+        zathura_image_t* image_it = girara_list_nth(images, idx);
+        zathura_rectangle_t* rect = g_try_malloc(sizeof(zathura_rectangle_t));
+        if (rect == NULL) {
+          break;
+        }
+        *rect = recalc_rectangle(page, image_it->position);
+        /* Scale rectangle coordinates by device factors to match surface pixel coordinates */
+        rect->x1 *= device_factors.x;
+        rect->x2 *= device_factors.x;
+        rect->y1 *= device_factors.y;
+        rect->y2 *= device_factors.y;
+        girara_list_append(rectangles, rect);
+      }
     }
   }
 
-  if (images != NULL) {
-    girara_list_free(images);
-  }
-  if (rectangles != NULL) {
-    girara_list_free(rectangles);
+  if (fast_formula == true) {
+    recolor_fast(priv, page_width, page_height, surface, rectangles, found_images);
+  } else {
+    recolor_slow(priv, page_width, page_height, surface, rectangles, found_images);
   }
 
   cairo_surface_mark_dirty(surface);
@@ -751,7 +864,7 @@ static bool render(render_job_t* job, ZathuraRenderRequest* request, ZathuraRend
 
   /* recolor */
   if (request_priv->render_plain == false && priv->recolor.enabled == true) {
-    recolor(priv, page, page_width, page_height, surface);
+    recolor(priv, page, page_width, page_height, surface, device_factors);
   }
 
   if (!invoke_completed_signal(job, surface)) {
@@ -787,18 +900,21 @@ static void render_job(void* data, void* user_data) {
 }
 
 void render_all(zathura_t* zathura) {
-  if (zathura == NULL || zathura->document == NULL) {
+  zathura_document_t* document = zathura_get_document(zathura);
+  if (document == NULL) {
     return;
   }
 
+  zathura_document_widget_compute_layout(ZATHURA_DOCUMENT_WIDGET(zathura->ui.document_widget));
+
   /* unmark all pages */
-  const unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
+  const unsigned int number_of_pages = zathura_document_get_number_of_pages(document);
   for (unsigned int page_id = 0; page_id < number_of_pages; ++page_id) {
-    zathura_page_t* page     = zathura_document_get_page(zathura->document, page_id);
+    zathura_page_t* page     = zathura_document_get_page(document, page_id);
     unsigned int page_height = 0, page_width = 0;
     const double height = zathura_page_get_height(page);
     const double width  = zathura_page_get_width(page);
-    page_calc_height_width(zathura->document, height, width, &page_height, &page_width, true);
+    page_calc_height_width(document, height, width, &page_height, &page_width, true);
 
     girara_debug("Queuing resize for page %u to %u x %u (%0.2f x %0.2f).", page_id, page_width, page_height, width,
                  height);
